from ..compat import string_type
from ..exceptions import InvalidTransmuteDefinition
from .response_type import ResponseType

class TransmuteAttributes(object):

    def __init__(self, paths=None, methods=None,
                 query_parameters=None, body_parameters=None,
                 header_parameters=None, path_parameters=None,
                 error_exceptions=None, response_types=None,
                 success_code=200):
        self.paths = set(paths or [])
        self.methods = set(methods or ["GET"])
        self.success_code = success_code
        self.query_parameters = set(query_parameters or [])
        self.body_parameters = self._coerce_parameters(body_parameters)
        self.header_parameters = set(header_parameters or [])
        self.path_parameters = set(path_parameters or [])
        self.error_exceptions = set(error_exceptions or [])
        self.response_types = {}
        for code, response in (response_types or {}).items():
            if not isinstance(response, ResponseType):
                response = ResponseType(**response)
            self.response_types[code] = response

    @staticmethod
    def _coerce_parameters(param):
        if isinstance(param, string_type):
            return param
        elif isinstance(param, set):
            return param
        elif isinstance(param, list):
            return set(param)
        elif isinstance(param, type(None)):
            return set()
        else:
            raise InvalidTransmuteDefinition("parameter must be a string, list or set. found: {0}".format(
                param
            ))

    def __or__(self, other):
        """
        merge values from another transmute function, taking the
        union of the two sets.
        """
        response_types = self.response_types.copy()
        for k, v in other.response_types.items():
            response_types[k] = v

        return TransmuteAttributes(
            paths=(self.paths | other.paths),
            methods=(self.methods | other.methods),
            success_code=(other.success_code or self.success_code),
            query_parameters=(self.query_parameters | other.query_parameters),
            body_parameters=self._join_parameters(other.body_parameters, self.body_parameters),
            header_parameters=(self.header_parameters | other.header_parameters),
            path_parameters=(self.path_parameters | other.path_parameters),
            error_exceptions=(self.error_exceptions | other.error_exceptions),
            response_types=response_types
        )

    def __str__(self):
        arg_list = []
        for k in ["paths", "methods",
                  "query_parameters", "body_parameters",
                  "header_parameters", "path_parameters",
                  "error_exceptions", "response_types",
                  "success_code"]:
            arg_list.append("{0}={1}".format(
                k, getattr(self, k)
            ))
        return "<TransmuteAttributes {0}>".format(" ".join(arg_list))

    @staticmethod
    def _join_parameters(base, nxt):
        """ join parameters from the lhs to the rhs, if compatible. """
        if nxt is None:
            return base
        if isinstance(base, set) and isinstance(nxt, set):
            return base | nxt
        else:
            return nxt



from .schematics_serializer import SchematicsSerializer
from .compound_serializer import (
    ListSerializer
)
from .interface import ObjectSerializer
from .serializer_set import ObjectSerializerSet
from .primitive_serializer import (
    BoolSerializer,
    FloatSerializer,
    IntSerializer,
    StringSerializer
)


def get_default_object_serializer_set():
    s = ObjectSerializerSet([
        BoolSerializer(),
        FloatSerializer(),
        IntSerializer(),
        StringSerializer(),
        SchematicsSerializer()
    ])
    s.serializers.append(ListSerializer(s))
    return s



import json
import pytest
from transmute_core.handler import process_result
from transmute_core import default_context, APIException, Response

CONTENT_TYPE = "application/json"


class SomeOtherException(Exception):
    pass


def test_process_result_200(complex_transmute_func):
    result = {"kind": "dog", "age": 5}
    exc = None
    output = process_result(
        complex_transmute_func,
        default_context, result,
        exc, CONTENT_TYPE
    )
    assert json.loads(output["body"].decode()) == result
    assert output["code"] == 200
    assert output["content-type"] == CONTENT_TYPE
    assert output["headers"] == {}


def test_process_result_api_exception(complex_transmute_func):
    result = None
    exc = APIException("foo")
    output = process_result(
        complex_transmute_func,
        default_context, result,
        exc, CONTENT_TYPE
    )
    assert json.loads(output["body"].decode()) == {
        "result": "invalid api use: " + str(exc),
        "success": False,
        "code": 400,
        "headers": {},
    }
    assert output["code"] == 400
    assert output["content-type"] == CONTENT_TYPE


def test_process_general_exception(complex_transmute_func):
    result = {"kind": "dog", "age": 5}
    exc = SomeOtherException("foo")
    with pytest.raises(SomeOtherException):
        process_result(
            complex_transmute_func,
            default_context, result,
            exc, CONTENT_TYPE
        )


def test_process_custom_code(transmute_func_custom_code):
    output = process_result(
        transmute_func_custom_code,
        default_context, 20,
        None, CONTENT_TYPE
    )
    assert output["code"] == 201


def test_process_result_multiple_types(response_transmute_func):
    result = Response("foo", 401)
    output = process_result(
        response_transmute_func,
        default_context, result,
        None, CONTENT_TYPE
    )
    assert json.loads(output["body"].decode()) == "foo"
    assert output["code"] == 401

    result = Response(False, 201)
    output = process_result(
        response_transmute_func,
        default_context, result,
        None, CONTENT_TYPE
    )
    assert json.loads(output["body"].decode()) is False
    assert output["code"] == 201


@pytest.mark.parametrize("content_type", [
    "application/myson",
    None,
])
def test_unknown_content_type_defaults_to_json(content_type,
                                               complex_transmute_func):
    result = {"kind": "dog", "age": 5}
    exc = None
    output = process_result(
        complex_transmute_func,
        default_context, result,
        exc, content_type
    )
    assert json.loads(output["body"].decode()) == result
    assert output["content-type"] == "application/json"



from ..compat import all_string_types, string_type


class IntSerializer:

    def can_handle(self, cls):
        return issubclass(cls, int)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "integer"}

    @staticmethod
    def load(cls, obj):
        return int(obj)

    @staticmethod
    def dump(cls, obj):
        return obj


class FloatSerializer:

    def can_handle(self, cls):
        return issubclass(cls, float)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "number"}

    @staticmethod
    def load(cls, obj):
        return float(obj)

    @staticmethod
    def dump(cls, obj):
        return obj


class StringSerializer:

    def can_handle(self, cls):
        return any(
            issubclass(cls, t) for t in all_string_types
        )

    @staticmethod
    def to_json_schema(cls):
        return {"type": "string"}

    @staticmethod
    def load(cls, obj):
        return obj

    @staticmethod
    def dump(cls, obj):
        return obj


class BoolSerializer:

    def can_handle(self, cls):
        return issubclass(cls, bool)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "boolean"}

    @staticmethod
    def load(cls, obj):
        if isinstance(obj, string_type):
            return obj.lower().startswith("t")
        return obj

    @staticmethod
    def dump(cls, obj):
        return obj



class RequestAdapter(object):
    """
    RequestAdapters bridge transmute's
    representation of a request, with the framework's
    implementation.

    implement the unimplemented methods.
    """

    @property
    def body(self):
        """ return the request body. """
        raise NotImplementedError()

    def _get_framework_args(self):
        """
        often, a framework provides specific variables that are passed
        into the handler function (e.g. the request object in
        aiohttp). return a dictionary of these arguments, which will be
        added to the function arguments if they appear.
        """
        raise NotImplementedError()

    def _query_argument(self, key, is_list):
        raise NotImplementedError()

    def _header_argument(self, key):
        raise NotImplementedError()

    def _path_argument(self, key):
        raise NotImplementedError()



import pytest
from transmute_core.object_serializers import (
    ListSerializer
)


@pytest.fixture
def list_serializer(object_serializer_set):
    return ListSerializer(object_serializer_set)


def test_list_serializer(list_serializer):
    x = ["10"]

    obj = list_serializer.load([int], x)
    assert obj == [10]
    obj = list_serializer.dump([int], obj)
    assert obj == [10]


def test_list_serializer_json_schema(list_serializer):
    schema = list_serializer.to_json_schema([str])
    assert schema == {
        "type": "array",
        "items": {"type": "string"}
    }


def test_list_serializer_can_handle(list_serializer):
    schema = list_serializer.to_json_schema([str])
    assert schema == {
        "type": "array",
        "items": {"type": "string"}
    }



import pytest
from transmute_core.compat import string_type
from transmute_core.exceptions import SerializationException
from datetime import datetime

NOW = datetime.now()


@pytest.mark.parametrize("inp, expected_output", [
    ("10", 10), ("-1", -1)
])
def test_int_load_happy(serializer, inp, expected_output):
    """ test all happy cases for the integer serializer """
    assert serializer.load(int, inp) == expected_output


@pytest.mark.parametrize("unhappy_input", [
    "foo", "bar"
])
def test_int_load_unhappy(serializer, unhappy_input):
    """ test all unhappy cases for the integer serializer """
    with pytest.raises(SerializationException):
        serializer.load(int, unhappy_input)


@pytest.mark.parametrize("inp, expected_output", [
    ("10", 10), ("1.0", 1.0)
])
def test_float_load_happy(serializer, inp, expected_output):
    """ test all happy cases for the integer serializer """
    assert serializer.load(float, inp) == expected_output


@pytest.mark.parametrize("unhappy_input", [
    "foo", "bar"
])
def test_float_load_unhappy(serializer, unhappy_input):
    """ test all unhappy cases for the integer serializer """
    with pytest.raises(SerializationException):
        serializer.load(float, unhappy_input)


@pytest.mark.parametrize("inp, expected_output", [
    ("true", True), ("false", False),
    (True, True), (False, False),
])
def test_bool_load_happy(serializer, inp, expected_output):
    assert serializer.load(bool, inp) is expected_output


@pytest.mark.parametrize("inp, expected_output", [
    ("foo", "foo")
])
def test_string_load_happy(serializer, inp, expected_output):
    assert serializer.load(string_type, inp) == expected_output


@pytest.mark.parametrize("inp, expected_output", [
    (NOW.isoformat(), NOW)
])
def test_datetime_load_happy(serializer, inp, expected_output):
    assert serializer.load(datetime, inp) == expected_output


@pytest.mark.parametrize("inp", [
    ("")
])
def test_datetime_load_unhappy(serializer, inp):
    with pytest.raises(SerializationException):
        serializer.load(datetime, inp)



import attr
from ..compat import string_type
from ..exceptions import InvalidTransmuteDefinition


@attr.s
class Param(object):
    """ a single parameter object. """
    argument_name = attr.ib()
    description = attr.ib(default=None)
    # the actual argument object. this gets populated later.
    arginfo = attr.ib(default=None)

    def __or__(self, other):
        raise InvalidTransmuteDefinition(
            "unable to join a single parameter with any other object."
        )

    def values(self):
        return (self,)

    def __setitem__(self, key, value):
        raise InvalidTransmuteDefinition(
            "unable to add parameter {0} to a single parameter.".format(key)
        )


class ParamSet(dict):
    """
    a param set represents a set of a particular grouping of
    parameters, such as header / query / body / etc.
    """

    def __or__(self, other):
        if not isinstance(other, ParamSet):
            raise TypeError("cannot join ParamSet with a {0}".format(type(other)))
        for name, definition in other.items():
            self[name] = definition



import swagger_schema
from transmute_core import default_context
from transmute_core.swagger import (
    generate_swagger_html, get_swagger_static_root,
    SwaggerSpec
)


def test_generate_swagger():
    swagger_static_root = "/statics/_swagger"
    swagger_json_url = "/swagger.json"
    body = generate_swagger_html(swagger_static_root, swagger_json_url)
    assert swagger_static_root in body
    assert swagger_json_url in body


def test_get_swagger_static_root():
    assert "static" in get_swagger_static_root()


def test_swagger_definition_generation():
    """
    swagger routes should be ablo to generate a proper
    spec.
    """
    routes = SwaggerSpec()
    assert routes.swagger_definition() == {
        "info": {"title": "example", "version": "1.0"},
        "paths": {},
        "swagger": "2.0",
        "basePath": "/"
    }


def test_swagger_transmute_func(transmute_func):
    """
    swagger routes should be ablo to generate a proper
    spec.
    """
    routes = SwaggerSpec()
    routes.add_func(transmute_func, default_context)
    assert routes.swagger_definition() == {
        "info": {"title": "example", "version": "1.0"},
        "paths": {
            "/api/v1/multiply": transmute_func.get_swagger_path(default_context).to_primitive(),
        },
        "swagger": "2.0",
        "basePath": "/"
    }


def test_multiple_response_types(response_transmute_func):
    """ multiple response types should be indicated as such in
    the swagger documentation.
    """
    routes = SwaggerSpec()
    routes.add_func(response_transmute_func, default_context)
    definition = routes.swagger_definition()
    path = definition["paths"]["/api/v1/create_if_authorized/"]
    responses = path["get"]["responses"]
    assert responses["201"] == swagger_schema.Response({
        "description": "",
        "schema": {"type": "boolean"},
        "headers": {
            "location": {"type": "string"}
        }
    }).to_primitive()
    assert responses["401"] == swagger_schema.Response({
        "description": "unauthorized",
        "schema": {"type": "string"}
    }).to_primitive()
    assert "200" not in responses
    assert "400" in responses


def test_swagger_add_path(transmute_func):
    """
    add_path should add the specified path to the main swagger object.
    """
    routes = SwaggerSpec()
    swagger_path = transmute_func.get_swagger_path(default_context)
    for p in transmute_func.paths:
        routes.add_path(p, swagger_path)
    assert routes.swagger_definition() == {
        "info": {"title": "example", "version": "1.0"},
        "paths": {
            "/api/v1/multiply": transmute_func.get_swagger_path(default_context).to_primitive(),
        },
        "swagger": "2.0",
        "basePath": "/"
    }


def test_basepath_override():
    assert SwaggerSpec().swagger_definition(base_path="dummy_path") == {
        'basePath': 'dummy_path',
        'info': {'title': 'example', 'version': '1.0'},
        'paths': {},
        'swagger': '2.0'
    }


def test_swagger_get_post(transmute_func, transmute_func_post):
    """
    adding different paths of diffrent methods should have both
    present in the spec.
    """
    routes = SwaggerSpec()
    routes.add_func(transmute_func, default_context)
    routes.add_func(transmute_func_post, default_context)
    spec = routes.swagger_definition()
    assert "get" in spec["paths"]["/api/v1/multiply"]
    assert "post" in spec["paths"]["/api/v1/multiply"]



import cProfile
import json
import pytest
import sys
from transmute_core import (
    TransmuteFunction, describe, annotate,
    ParamExtractor, NoArgument, process_result
)
from schematics.models import Model
from schematics.types import StringType, BooleanType, IntType
from schematics.types.compound import ModelType


class User(Model):
    name = StringType()
    age = IntType()


class ComplexModel(Model):
    user = ModelType(User)
    description = StringType()
    is_allowed = BooleanType()


@describe(paths="/foo", body_parameters="body")
@annotate({"body": ComplexModel, "return": ComplexModel})
def complex_body_method(body):
    return body


@describe(paths="/foo", body_parameters="body")
@annotate({"body": int, "return": int})
def simple_body_method(body):
    return body


@describe(paths="/foo", body_parameters="body")
@annotate({"body": str, "return": str})
def body_string(body):
    return body


def execute(context, func, obj_as_json):
    extractor = ParamExtractorMock(obj_as_json)
    args, kwargs = extractor.extract_params(
        context, func, "application/json"
    )
    exc, result = None, None
    try:
        result = func(*args, **kwargs)
    except Exception as e:
        exc = e
        exc.__traceback__ = sys.exc_info[:2]
    process_result(func, context, result, exc, "application/json")


def test_large_str_benchmark(benchmark, context):
    """
    a benchmark of a fake full execution flow of a transmute function.
    """
    s = "a" * 100000

    func = TransmuteFunction(body_string)
    obj_json = json.dumps(s)

    benchmark(lambda: execute(context, func, obj_json))


def test_complex_benchmark(benchmark, context):
    """
    a benchmark of a fake full execution flow of a transmute function.
    """
    obj = ComplexModel({
        "user": {
            "name": "Richard Stallman",
            "age": 104
        },
        "description": "this is a test",
        "is_allowed": True
    })

    complex_func = TransmuteFunction(complex_body_method)
    complex_json = json.dumps(obj.to_primitive())

    benchmark(lambda: execute(context, complex_func, complex_json))


def test_simple_benchmark(benchmark, context):

    simple_func = TransmuteFunction(simple_body_method)
    simple_json = json.dumps(1)

    benchmark(lambda: execute(context, simple_func, simple_json))

    # def profile():
    #     for i in range(10000):
    #         execute()
    # cProfile.runctx('profile()', globals(), locals())


class ParamExtractorMock(ParamExtractor):

    def __init__(self, body):
        self._body = body

    @property
    def body(self):
        return self._body

    def _get_framework_args(self):
        return {}

    def _query_argument(self, key, is_list):
        return NoArgument

    def _header_argument(self, key):
        return NoArgument

    def _path_argument(self, key):
        return NoArgument



from transmute_core.http_parameters import (
    get_swagger_parameters, ParamSet, Param, Parameters
)
from transmute_core.function.signature import Argument, NoDefault


def test_post_schema_swagger(parameter_post_schema, context):
    assert get_swagger_parameters(parameter_post_schema, context)[0].to_primitive() == {
        "in": "body",
        "name": "body",
        "required": True,
        "schema": {
            "type": "integer"
        }
    }


def test_header_only_schema(context):
    params = Parameters(
        path=ParamSet({"path": Param(argument_name="path",
                                     arginfo=Argument("path", NoDefault, int))}),
    )
    assert get_swagger_parameters(params, context)[0].to_primitive() == {
        "in": "path",
        "name": "path",
        "required": True,
        "type": "integer"
    }



import pytest
from transmute_core import annotate, describe
from transmute_core.compat import getfullargspec
from transmute_core.function.signature import (
    FunctionSignature,
)
from transmute_core.function.parameters import (
    get_parameters, _extract_path_parameters_from_paths
)
from transmute_core.attributes import TransmuteAttributes
from transmute_core.exceptions import InvalidTransmuteDefinition


def test_get_argument_set():

    @describe(paths="/{x}", body_parameters=["y"])
    @annotate({"x": int, "y": float, "z": int, "width": int, "height": float})
    def make_square(x, y, z, width=None, height=12):
        pass

    argspec = getfullargspec(make_square)
    signature = FunctionSignature.from_argspec(argspec)
    argument_sets = get_parameters(signature, make_square.transmute)
    assert ["y"] == list(argument_sets.body.keys())
    assert set(["x"]) == set(argument_sets.path.keys())
    assert set(["z", "width", "height"]) == set(argument_sets.query.keys())


def test_ignore_request_parameter():

    @annotate({"x": int, "y": int})
    def handle_request(request, x, y):
        pass

    argspec = getfullargspec(handle_request)
    signature = FunctionSignature.from_argspec(argspec)
    params = get_parameters(signature, TransmuteAttributes(),
                            arguments_to_ignore=["request"])
    for typ in ["query", "body", "header", "path"]:
        assert "request" not in getattr(params, typ)


@pytest.mark.parametrize("invalid_paths, expected", [
    (["/foo/{bar}/baz", "/foo/{baz}/bar"], {"bar", "baz"}),
    (["/foo/{bar}/baz", "/foo/{bar}/bar/{extra}"], {"bar", "extra"}),
])
def test_non_matching_path_parameters(invalid_paths, expected):
    """ with non matching paths, the whole set of values should be used. """
    assert _extract_path_parameters_from_paths(invalid_paths) == expected


@pytest.mark.parametrize("paths, params", [
    ([], [])
])
def test_param_list(paths, params):
    assert _extract_path_parameters_from_paths(paths) == set(params)



class ContentTypeSerializer(object):
    """
    A ContentTypeSerializer handles the conversion from
    a python data structure to a bytes object representing
    the content in a particular content type.
    """

    def content_type(self):
        """
        return back what a list of content types
        this serializer should support.
        """
        raise NotImplementedError()  # noqa

    def main_type(self):
        """
        return back a single content type that represents this
        serializer.
        """
        raise NotImplementedError()  # noqa

    def dump(data):
        """
        should return back a bytes (or string in python 2),
        representation of your object, to be used in e.g. response
        bodies.

        a ValueError should be returned in the case where
        the object cannote be serialized.
        """
        raise NotImplementedError()  # noqa

    def load(raw_bytes):
        """
        given a bytes object, should return a base python data
        structure that represents the object.

        a ValueError should be returned in the case where
        the object cannot be serialized.
        """
        raise NotImplementedError()  # noqa

    def can_handle(content_type_name):
        """
        given a content type, returns true if this serializer
        can convert bodies of the given type.
        """
        raise NotImplementedError()  # noqa



import attr
import copy
from ..compat import string_type
from swagger_schema import Response

# @attr.s
# class ParameterType(object):
#     description =

@attr.s
class ResponseType(object):
    type = attr.ib()
    description = attr.ib(
        validator=[
            attr.validators.instance_of(string_type)
        ],
        default=""
    )
    headers = attr.ib(
        validator=[
            attr.validators.instance_of(dict)
        ],
        default=attr.Factory(dict)
    )

    def swagger_definition(self, context):
        if self.type is not None:
            type_definition = context.serializers.to_json_schema(self.type)
            schema = context.response_shape.swagger(copy.deepcopy(type_definition))
        else:
            schema = {"type": "object"}
        headers = {}

        for name, header in self.headers.items():
            header_definition = context.serializers.to_json_schema(header["type"])
            if "description" in header:
                header_definition["description"] = header["description"]
            headers[name] = header_definition

        response = {
            "description": self.description,
            "schema": schema
        }

        if headers:
            response["headers"] = headers

        return Response(response)



"""
utilities to help with generating handlers.
"""
import attr
from .exceptions import (
    APIException, NoSerializerFound
)
from .function import Response
from six import reraise


def process_result(transmute_func, context, result, exc, content_type):
    """
    process a result:

    transmute_func: the transmute_func function that returned the response.

    context: the transmute_context to use.

    result: the return value of the function, which will be serialized and
            returned back in the API.

    exc: the exception object. For Python 2, the traceback should
         be attached via the __traceback__ attribute. This is done automatically
         in Python 3.

    content_type: the content type that request is requesting for a return type.
                  (e.g. application/json)
    """
    if isinstance(result, Response):
        response = result
    else:
        response = Response(
            result=result,
            code=transmute_func.success_code,
            success=True
        )
    if exc:
        if isinstance(exc, APIException):
            response.result = "invalid api use: {0}".format(str(exc))
            response.success = False
            response.code = exc.code
        else:
            reraise(type(exc), exc, getattr(exc, "__traceback__", None))
    else:
        return_type = transmute_func.get_response_by_code(response.code)
        if return_type:
            response.result = context.serializers.dump(return_type, response.result)
    try:
        content_type = str(content_type)
        serializer = context.contenttype_serializers[content_type]
    except NoSerializerFound:
        serializer = context.contenttype_serializers.default
        content_type = serializer.main_type
    if response.success:
        result = context.response_shape.create_body(attr.asdict(response))
        response.result = result
    else:
        response.result = attr.asdict(response)
    body = serializer.dump(response.result)
    # keeping the return type a dict to
    # reduce performance overhead.
    return {
        "body": body,
        "code": response.code,
        "content-type": content_type,
        "headers": response.headers
    }



class ListSerializer(object):
    """ serializer for lists """

    def __init__(self, subtype_serializer):
        self._subtype_serializer = subtype_serializer

    def can_handle(self, cls):
        if isinstance(cls, list):
            return True

    def to_json_schema(self, cls):
        subtype = cls[0]
        subtype_serializer = self._subtype_serializer[subtype]
        subtype_schema = subtype_serializer.to_json_schema(subtype)
        return {
            "type": "array",
            "items": subtype_schema
        }

    def dump(self, cls, value):
        subtype = cls[0]
        serializer = self._subtype_serializer[subtype]
        return [
            serializer.dump(subtype, el) for el in value
        ]

    def load(self, cls, value):
        subtype = cls[0]
        serializer = self._subtype_serializer[subtype]
        return [
            serializer.load(subtype, el) for el in value
        ]



import re
from ..http_parameters import Parameters, ParamSet, Param
from ..exceptions import InvalidTransmuteDefinition


def get_parameters(signature, transmute_attrs, arguments_to_ignore=None):
    """given a function, categorize which arguments should be passed by
    what types of parameters. The choices are:

    * query parameters: passed in as query parameters in the url
    * body parameters: retrieved from the request body (includes forms)
    * header parameters: retrieved from the request header
    * path parameters: retrieved from the uri path

    The categorization is performed for an argument "arg" by:

    1. examining the transmute parameters attached to the function (e.g.
    func.transmute_query_parameters), and checking if "arg" is mentioned. If so,
    it is added to the category.

    2. If the argument is available in the path, it will be added
    as a path parameter.

    3. If the method of the function is GET and only GET, then "arg" will be
    be added to the expected query parameters. Otherwise, "arg" will be added as
    a body parameter.
    """
    params = Parameters()
    used_keys = set(arguments_to_ignore or [])
    # examine what variables are categorized first.
    for key in ["query", "header", "path"]:
        param_set = getattr(params, key)
        explicit_parameters = getattr(transmute_attrs, key + "_parameters")
        used_keys |= load_parameters(param_set, explicit_parameters, signature)

    body_parameters = transmute_attrs.body_parameters
    if isinstance(body_parameters, str):
        name = body_parameters
        params.body = Param(argument_name=name,
                            arginfo=signature.get_argument(name))
        used_keys.add(name)
    else:
        used_keys |= load_parameters(params.body, transmute_attrs.body_parameters,
                                     signature)

    # extract the parameters from the paths
    for name in _extract_path_parameters_from_paths(transmute_attrs.paths):
        params.path[name] = Param(argument_name=name, arginfo=signature.get_argument(name))
        used_keys.add(name)

    # check the method type, and decide if the parameters should be extracted
    # from query parameters or the body
    default_param_key = (
        "query" if transmute_attrs.methods == set(["GET"]) else "body"
    )
    default_params = getattr(params, default_param_key)

    # parse all positional params
    for arginfo in signature:
        if arginfo.name in used_keys:
            continue
        used_keys.add(arginfo.name)
        default_params[arginfo.name] = Param(arginfo.name, arginfo=arginfo)

    return params

PART_REGEX = re.compile(r'(\{[_a-zA-Z][^{}]*(?:\{[^{}]*\}[^{}]*)*\})')
PARAM_REGEX = re.compile(r'^\{(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)\}$')


def _extract_path_parameters_from_paths(paths):
    """
    from a list of paths, return back a list of the
    arguments present in those paths.

    the arguments available in all of the paths must match: if not,
    an exception will be raised.
    """
    params = set()
    for path in paths:
        parts = PART_REGEX.split(path)
        for p in parts:
            match = PARAM_REGEX.match(p)
            if match:
                params.add(match.group("name"))
    return params


def load_parameters(param_set, param_list, signature):
    used_keys = set()
    for name in param_list:
        param_set[name] = Param(argument_name=name,
                                arginfo=signature.get_argument(name))
        used_keys.add(name)
    return used_keys



from transmute_core.exceptions import NoSerializerFound


class ObjectSerializerSet(object):
    """
    composes multiple serializers, delegating commands to one
    that can handle the desired content type.

    SerializerSet implements a dict-like interface. Retrieving
    serializers is done by get the content type item:

    .. code-block:: python

        serializers["application/json"]
    """

    def __init__(self, serializer_list):
        self.serializers = serializer_list
        self._cache = {}
        assert len(self.serializers) > 0, "at least one serializer should be passed!"

    def _get_serializer_for_type(self, cls):
        cls_key = self._to_key(cls)
        if cls_key not in self._cache:
            for serializer in self.serializers:
                try:
                    if serializer.can_handle(cls):
                        self._cache[cls_key] = serializer
                        return serializer
                except:
                    pass
            raise NoSerializerFound(
                "unable to find serializer for " + str(cls) + " in: " +
                str(self.serializers)
            )
        return self._cache[cls_key]

    def __getitem__(self, cls):
        """
        from a given content type, returns the appropriate serializer.

        the list of serializers are interated, from first added to last,
        and returns the one who's can_handle returns true.
        """
        return self._get_serializer_for_type(cls)

    def to_json_schema(self, cls):
        return self[cls].to_json_schema(cls)

    def dump(self, cls, value):
        return self[cls].dump(cls, value)

    def load(self, cls, value):
        return self[cls].load(cls, value)

    @staticmethod
    def _to_key(target_cls):
        if isinstance(target_cls, list):
            return tuple(target_cls)
        return target_cls



def test_hashing_list_object_serializer(object_serializer_set):
    """
    a the object serializer set should supported
    a list object.
    """
    obj = [str]
    object_serializer_set[obj]
    object_serializer_set[obj]



import uuid
import pytest
from datetime import datetime
from schematics.models import Model
from schematics.types import StringType, IntType, BaseType, UUIDType, URLType, Serializable
from schematics.types.compound import DictType
from schematics.exceptions import ValidationError
from transmute_core.exceptions import SerializationException


class Card(Model):
    name = StringType()
    price = IntType()


def greater_than_zero(value):
    if value < 0:
        raise ValidationError("must be greater than zero")
    return value


class Person(Model):
    age = IntType(required=True, validators=[greater_than_zero])
    bio = StringType()

card_dict = {"name": "foo", "price": 10}
card = Card(card_dict)


@pytest.mark.parametrize("typ,inp,out", [
    (Card, card, card_dict),
    ([Card], [card], [card_dict])
])
def test_schematics_integration_dump(serializer, typ, inp, out):
    assert serializer.dump(typ, inp) == out


def test_schematics_validate_is_called(serializer):
    with pytest.raises(SerializationException):
        serializer.load(Person, {"age": -1, "bio": "foo"})

now = datetime.now()


@pytest.mark.parametrize("typ,inp,out", [
    (Card, card_dict, card),
    ([Card], [card_dict], [card]),
    (datetime, now, now)
])
def test_schematics_integration_load(serializer, typ, inp, out):
    assert serializer.load(typ, inp) == out


def test_schematics_to_json_schema(serializer):
    assert serializer.to_json_schema(Card) == {
        "properties": {
            "name": {"type": "string"},
            "price": {"type": "number"}
        },
        "title": "Card",
        "type": "object"
    }


@pytest.mark.parametrize("inp, expected", [
    (UUIDType, {"type": "string", "format": "uuid"}),
    (URLType, {"type": "string", "format": "url"}),
    (StringType, {"type": "string"}),
    (BaseType, {"type": "object"}),
    (datetime, {"type": "string", "format": "date-time"}),
    (Serializable(fget=lambda: None, type=StringType()), {"type": "string"}),
])
def test_to_json_schema(serializer, inp, expected):
    assert serializer.to_json_schema(inp) == expected


def test_schematics_to_json_schema_required_value(serializer):

    class CardNameRequired(Model):
        name = StringType(required=True)
        price = IntType()

    assert serializer.to_json_schema(CardNameRequired) == {
        "properties": {
            "name": {"type": "string"},
            "price": {"type": "number"}
        },
        "title": "CardNameRequired",
        "required": ["name"],
        "type": "object"
    }


def test_type_not_found_defaults_to_object(serializer):

    class MyType(BaseType):
        pass

    assert serializer.to_json_schema(MyType()) == {
        "type": "object"
    }


def test_non_model_or_primitive_raises_exception(serializer):

    class MyType(object):
        pass

    with pytest.raises(SerializationException):
        serializer.to_json_schema(MyType)


def test_to_json_list(serializer):
    assert serializer.to_json_schema([int]) == {
        "type": "array",
        "items": {"type": "number"}
    }


def test_to_json_dict(serializer):
    assert serializer.to_json_schema(DictType(IntType())) == {
        "type": "object",
        "additionalProperties": {"type": "number"}
    }


def test_serialize_type(serializer):
    """ serializer should be able to serialize a raw type. """
    assert serializer.dump(DictType(StringType()), {"key": "foo"}) == {"key": "foo"}


def test_uuid_serializer(serializer):
    i = uuid.uuid4()
    dumped = serializer.dump(UUIDType, i)
    assert dumped == str(i)
    assert serializer.load(UUIDType, dumped) == i



from swagger_schema import (
    BodyParameter, QueryParameter, HeaderParameter, PathParameter,
)
from ..function.signature import NoDefault
from .param_set import Param, ParamSet


def get_swagger_parameters(parameters, context):
    ret_parameters = []
    for name, param in parameters.query.items():
        arginfo = param.arginfo
        ret_parameters.append(QueryParameter({
            "name": name,
            "required": arginfo.default is NoDefault,
            "type": context.serializers.to_json_schema(arginfo.type)["type"],
        }))

    for name, param in parameters.header.items():
        arginfo = param.arginfo
        ret_parameters.append(HeaderParameter({
            "name": name,
            "required": arginfo.default is NoDefault,
            "type": context.serializers.to_json_schema(arginfo.type)["type"],
        }))

    body_param = _build_body_schema(context.serializers, parameters.body)
    if body_param is not None:
        ret_parameters.append(body_param)

    for name, param in parameters.path.items():
        arginfo = param.arginfo
        ret_parameters.append(PathParameter({
            "name": name,
            "required": True,
            "type": context.serializers.to_json_schema(arginfo.type)["type"],
        }))

    return ret_parameters


def _build_body_schema(serializer, body_parameters):
    """ body is built differently, since it's a single argument no matter what. """
    if isinstance(body_parameters, Param):
        schema = serializer.to_json_schema(body_parameters.arginfo.type)
        required = True
    else:
        if len(body_parameters) == 0:
            return None
        required = set()
        body_properties = {}
        for name, param in body_parameters.items():
            arginfo = param.arginfo
            body_properties[name] = serializer.to_json_schema(arginfo.type)
            if arginfo.default is NoDefault:
                required.add(name)
        schema = {
            "type": "object",
            "required": list(required),
            "properties": body_properties
        }
        required = len(required) > 0
    return BodyParameter({
        "name": "body",
        "required": required,
        "schema": schema
    })



from .param_set import Param, ParamSet
from .swagger import get_swagger_parameters

class Parameters(object):
    def __init__(self, query=None, body=None, header=None, path=None):
        self.query = query or {}
        self.body = body or {}
        self.header = header or {}
        self.path = path or {}



from transmute_core.attributes import TransmuteAttributes, ResponseType
from schematics.types.net import URLType
from swagger_schema import Response


def test_merge():
    left = TransmuteAttributes(
        methods=["POST"], query_parameters=["a"]
    )
    right = TransmuteAttributes(
        methods=["PUT"], query_parameters=["b"],
        body_parameters=["c"]
    )
    joined = left | right
    assert joined.methods == set(["POST", "PUT"])
    assert joined.query_parameters == set(["a", "b"])
    assert joined.body_parameters == set(["c"])


def test_merge_response_type_by_code():
    left = TransmuteAttributes(response_types={
        200: {"type": bool}, 201: {"type": bool}
    })
    right = TransmuteAttributes(response_types={
        201: {"type": str}
    })
    joined = left | right
    assert joined.response_types == {
        200: ResponseType(type=bool), 201: ResponseType(type=str)
    }


def test_swagger_definition_if_none(context):
    """ a response type that is not specified, should result in a generic swagger schema. """
    response_type = ResponseType(type=None)
    definition = response_type.swagger_definition(context)
    assert definition.to_primitive() == {
        "description": "",
        "schema": {"type": "object"}
    }


def test_header_in_response():
    left = TransmuteAttributes(response_types={
        200: {
            "type": bool,
            "headers": {
                "location": {"type": URLType}
            }
        }
    })
    assert left.response_types == {
        200: ResponseType(type=bool, headers={
            "location": {"type": URLType}
        })
    }




def test_merge_response_success_code():
    left = TransmuteAttributes(success_code=200)
    right = TransmuteAttributes(success_code=201)
    joined = left | right
    assert joined.success_code == 201

def test_merge_body_parameters_argument():
    right = TransmuteAttributes()
    left = TransmuteAttributes(body_parameters="body")
    joined = left | right
    assert joined.body_parameters == "body"



import pytest
from transmute_core.http_parameters import Parameters, Param, ParamSet
from transmute_core.function.signature import Argument, NoDefault

@pytest.fixture
def parameters():
    return Parameters(
        query=ParamSet({"query": Param(argument_name="query",
                                       arginfo=Argument("query", None, int))}),
        body=ParamSet({
            "left": Param(argument_name="left",
                          arginfo=Argument("left", NoDefault, int)),
            "right": Param(argument_name="right",
                           arginfo=Argument("right", 2, int))
        }),
        header=ParamSet({"header": Param(argument_name="header",
                                         arginfo=Argument("header", NoDefault, int))}),
        path=ParamSet({"path": Param(argument_name="path",
                                     arginfo=Argument("path", NoDefault, int))}),
    )


@pytest.fixture
def parameter_post_schema():
    return Parameters(
        body=Param(argument_name="body",
                   arginfo=Argument("left", NoDefault, int)),
    )



from transmute_core.function.signature import Argument, NoDefault
from transmute_core.http_parameters import get_swagger_parameters
from transmute_core import default_context
from transmute_core.http_parameters import Parameters, Param, ParamSet


def test_swagger_parameters():
    parameters = Parameters(
        query=ParamSet({"query": Param(argument_name="query",
                                       arginfo=Argument("query", None, int))}),
        body=ParamSet({
            "left": Param(argument_name="left",
                          arginfo=Argument("left", NoDefault, int)),
            "right": Param(argument_name="right",
                           arginfo=Argument("right", 2, int))
        }),
        header=ParamSet({"header": Param(argument_name="header",
                                         arginfo=Argument("header", NoDefault, int))}),
        path=ParamSet({"path": Param(argument_name="path",
                                     arginfo=Argument("path", NoDefault, int))}),
    )
    params = get_swagger_parameters(parameters, default_context)
    params = [p.to_primitive() for p in params]
    assert {
        "in": "query",
        "name": "query",
        "required": False,
        "type": "integer"
    } in params
    assert {
        "in": "body",
        "name": "body",
        "schema": {
            "type": "object",
            "required": ["left"],
            "properties": {
                "left": {"type": "integer"},
                "right": {"type": "integer"}
            }
        },
        "required": True,
    } in params
    assert {
        "in": "header",
        "name": "header",
        "type": "integer",
        "required": True,
    } in params
    assert {
        "in": "path",
        "name": "path",
        "type": "integer",
        "required": True,
    } in params



import pytest
from transmute_core import NoSerializerFound, SerializationException


def test_default_serializer_json(serializer_set):
    frm, expected_to = {"foo": "bar"}, b'{"foo": "bar"}'
    serializer = serializer_set["application/json"]
    assert serializer.dump(frm) == expected_to
    assert serializer.load(expected_to) == frm


def test_default_serializer_yaml(serializer_set):
    frm, expected_to = {"foo": "bar"}, b'foo: bar\n'
    serializer = serializer_set["application/yaml"]
    assert serializer.dump(frm) == expected_to
    assert serializer.load(expected_to) == frm


def test_default_serializer_prop(serializer_set):
    assert serializer_set.default.main_type == "application/json"


def test_no_serializer_found_raises_exception(serializer_set):
    with pytest.raises(NoSerializerFound):
        assert serializer_set["oogabooga"]


@pytest.mark.parametrize("content_type,bad_input", [
    ("application/yaml", b"[a, !eafia']atedntad}"),
    ("application/json", b"{\"ooga")
])
def test_bad_object_raises_serialization_exception(serializer_set, content_type, bad_input):
    """ a bad object serialization should raise a serialization exception """
    with pytest.raises(SerializationException):
        serializer_set[content_type].load(bad_input)


def test_keys(serializer_set):
    assert serializer_set.keys() == [
        "application/json", "application/x-yaml"
    ]
